import typing

from modern_di.providers import AbstractProvider
from modern_di.scope import Scope


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class ContainerProvider(AbstractProvider[typing.Any]):
    __slots__ = AbstractProvider.BASE_SLOTS

    def __init__(self, scope: Scope) -> None:
        super().__init__(scope)



import typing

from modern_di.providers.abstract import AbstractCreatorProvider
from modern_di.scope import Scope


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class Factory(AbstractCreatorProvider[T_co]):
    __slots__ = AbstractCreatorProvider.BASE_SLOTS

    def __init__(
        self,
        scope: Scope,
        creator: typing.Callable[P, T_co],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        super().__init__(scope, creator, *args, **kwargs)



import typing

from modern_di.providers import AbstractProvider
from modern_di.scope import Scope


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class ContextProvider(AbstractProvider[T_co]):
    __slots__ = [*AbstractProvider.BASE_SLOTS, "context_type", "required"]

    def __init__(self, scope: Scope, context_type: type[T_co]) -> None:
        super().__init__(scope)
        self.context_type = context_type



import dataclasses
import typing
import warnings

from modern_di.providers.abstract import AbstractProvider


T_co = typing.TypeVar("T_co", covariant=True)


@dataclasses.dataclass(kw_only=True, slots=True, frozen=True)
class ProvidersRegistry:
    providers_by_name: dict[str, AbstractProvider[typing.Any]] = dataclasses.field(init=False, default_factory=dict)
    providers_by_type: dict[type, AbstractProvider[typing.Any]] = dataclasses.field(init=False, default_factory=dict)

    def find_provider(
        self, dependency_name: str | None = None, dependency_type: type[T_co] | None = None
    ) -> AbstractProvider[T_co] | None:
        if dependency_name and (provider := self.providers_by_name.get(dependency_name)):
            return provider

        if dependency_type and (provider := self.providers_by_type.get(dependency_type)):
            return provider

        return None

    def add_providers(self, **kwargs: AbstractProvider[typing.Any]) -> None:
        for provider_name, provider in kwargs.items():
            if provider_name in self.providers_by_name:
                warnings.warn(
                    f"Provider is duplicated by name {provider_name}. Choose unique name if you want resolving by name",
                    RuntimeWarning,
                    stacklevel=2,
                )
                del self.providers_by_name[provider_name]
            else:
                self.providers_by_name[provider_name] = provider

            provider_type = provider.bound_type
            if not provider_type:
                continue

            if provider_type in self.providers_by_type:
                warnings.warn(
                    f"Provider is duplicated by type {provider_type}. "
                    f"Bind provider to unique type if you want resolving by type",
                    RuntimeWarning,
                    stacklevel=2,
                )
                del self.providers_by_type[provider_type]
                continue

            self.providers_by_type[provider_type] = provider



import types
import typing


def define_bound_type(creator: type | object) -> type | None:
    if isinstance(creator, type):
        return creator

    type_hints = typing.get_type_hints(creator)
    return_annotation = type_hints.get("return")
    if not return_annotation:
        return None

    if isinstance(return_annotation, type) and not isinstance(return_annotation, types.GenericAlias):
        return return_annotation

    return None



import typing

from modern_di.providers.abstract import AbstractProvider
from modern_di.scope import Scope


T_co = typing.TypeVar("T_co", covariant=True)


class Dict(AbstractProvider[dict[str, T_co]]):
    __slots__ = AbstractProvider.BASE_SLOTS

    def __init__(self, scope: Scope, **kwargs: AbstractProvider[T_co]) -> None:
        super().__init__(scope, kwargs=kwargs)



import abc
import typing
import uuid

import typing_extensions

from modern_di.helpers.type_helpers import define_bound_type
from modern_di.scope import Scope


T_co = typing.TypeVar("T_co", covariant=True)
R = typing.TypeVar("R")
P = typing.ParamSpec("P")


class AbstractProvider(abc.ABC, typing.Generic[T_co]):
    BASE_SLOTS: typing.ClassVar[list[str]] = ["scope", "provider_id", "args", "kwargs", "bound_type"]

    def __init__(
        self,
        scope: Scope,
        args: list[typing.Any] | None = None,
        kwargs: dict[str, typing.Any] | None = None,
        bound_type: type | None = None,
    ) -> None:
        self.scope = scope
        self.provider_id: typing.Final = str(uuid.uuid4())
        self.args = args
        self.kwargs = kwargs
        self.bound_type = bound_type
        self._check_providers_scope()

    def bind_type(self, new_type: type | None) -> typing_extensions.Self:
        self.bound_type = new_type
        return self

    @property
    def cast(self) -> T_co:
        return typing.cast(T_co, self)

    def _check_providers_scope(self) -> None:
        if self.args:
            for provider in self.args:
                if isinstance(provider, AbstractProvider) and provider.scope > self.scope:
                    msg = f"Scope of dependency is {provider.scope.name} and current scope is {self.scope.name}"
                    raise RuntimeError(msg)

        if self.kwargs:
            for name, provider in self.kwargs.items():
                if isinstance(provider, AbstractProvider) and provider.scope > self.scope:
                    msg = f"Scope of {name} is {provider.scope.name} and current scope is {self.scope.name}"
                    raise RuntimeError(msg)


class AbstractCreatorProvider(AbstractProvider[T_co], abc.ABC):
    BASE_SLOTS: typing.ClassVar = [*AbstractProvider.BASE_SLOTS, "creator"]

    def __init__(
        self,
        scope: Scope,
        creator: typing.Callable[P, typing.Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        super().__init__(scope, args=list(args), kwargs=kwargs, bound_type=define_bound_type(creator))
        self.creator: typing.Final = creator



import typing

from modern_di.providers.abstract import AbstractProvider
from modern_di.scope import Scope


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class Object(AbstractProvider[T_co]):
    __slots__ = [*AbstractProvider.BASE_SLOTS, "obj"]

    def __init__(self, scope: Scope, obj: T_co) -> None:
        super().__init__(scope)
        self.obj: typing.Final = obj



import typing

from modern_di import providers


if typing.TYPE_CHECKING:
    from modern_di.container import Container


T_co = typing.TypeVar("T_co", covariant=True)


def _resolve_args(container: "Container", args: list[typing.Any]) -> list[typing.Any]:
    return [container.resolve_provider(x) if isinstance(x, providers.AbstractProvider) else x for x in args]


def _resolve_kwargs(container: "Container", kwargs: dict[str, typing.Any]) -> dict[str, typing.Any]:
    return {
        k: container.resolve_provider(v) if isinstance(v, providers.AbstractProvider) else v for k, v in kwargs.items()
    }


def _resolve_dict(container: "Container", provider: providers.Dict[T_co]) -> dict[str, T_co]:
    return _resolve_kwargs(container, provider.kwargs or {})


def _resolve_list(container: "Container", provider: providers.List[T_co]) -> list[T_co]:
    return _resolve_args(container, provider.args or [])


def _resolve_container(container: "Container", _: providers.ContainerProvider) -> "Container":
    return container


def _resolve_context(container: "Container", provider: providers.ContextProvider[T_co]) -> T_co | None:
    return container.context_registry.find_context(provider.context_type)


def _resolve_factory(container: "Container", provider: providers.Factory[T_co]) -> T_co:
    if (override := container.overrides_registry.fetch_override(provider.provider_id)) is not None:
        return typing.cast(T_co, override)

    args = _resolve_args(container, provider.args or [])
    kwargs = _resolve_kwargs(container, provider.kwargs or {})
    return typing.cast(T_co, provider.creator(*args, **kwargs))


def _resolve_singleton(container: "Container", provider: providers.Singleton[T_co]) -> T_co:
    if (override := container.overrides_registry.fetch_override(provider.provider_id)) is not None:
        return typing.cast(T_co, override)

    if (provider_state := container.state_registry.fetch_provider_state(provider)) is not None:
        return provider_state

    args = _resolve_args(container, provider.args or [])
    kwargs = _resolve_kwargs(container, provider.kwargs or {})

    if container.lock:
        container.lock.acquire()

    try:
        if (provider_state := container.state_registry.fetch_provider_state(provider)) is not None:
            return provider_state

        instance = typing.cast(T_co, provider.creator(*args, **kwargs))
        container.state_registry.set_provider_state(provider, instance)
        return instance
    finally:
        if container.lock:
            container.lock.release()


RESOLVERS: dict[type[providers.AbstractProvider[typing.Any]], typing.Callable[..., typing.Any]] = {
    providers.Dict: _resolve_dict,
    providers.List: _resolve_list,
    providers.ContainerProvider: _resolve_container,
    providers.ContextProvider: _resolve_context,
    providers.Factory: _resolve_factory,
    providers.Singleton: _resolve_singleton,
}



import dataclasses
import typing

from modern_di.providers import Singleton


T_co = typing.TypeVar("T_co", covariant=True)


@dataclasses.dataclass(kw_only=True, slots=True, frozen=True)
class StateRegistry:
    states: dict[str, typing.Any] = dataclasses.field(init=False, default_factory=dict)

    def fetch_provider_state(self, provider: Singleton[T_co]) -> T_co | None:
        return self.states.get(provider.provider_id)

    def set_provider_state(self, provider: Singleton[T_co], instance: typing.Any) -> None:  # noqa: ANN401
        self.states[provider.provider_id] = instance
